from .api import *  # noqa

# Explicit symbols export
__all__ = api_exports  # noqa

# Package metadata
__author__ = 'Tomas Aparicio'
__license__ = 'MIT'

# Current version
__version__ = '0.1.0-rc.1'
