class PookInvalidBody(Exception):
    pass


class PookNoMatches(Exception):
    pass


class PookNetworkFilterError(Exception):
    pass


class PookExpiredMock(Exception):
    pass


class PookInvalidArgument(Exception):
    pass
